Public Class SuperQwerty
    Inherits AppActionEngine

    Private Class ActionHugeText
        Inherits ActionWithText

        Private myText As String = Nothing
        Private myColor As Color = Nothing
        Private myTimeSpan As TimeSpan = Nothing

        Public Sub New(ByVal unreg As UnregisterActionCallback, ByVal bounds As Rectangle, ByVal f As Font, ByVal t As String, ByVal c As Color, ByVal ts As TimeSpan)
            MyBase.New(unreg, bounds, f)

            myText = t
            myColor = c
            myTimeSpan = ts

        End Sub

        Protected Overrides Sub TextOut()

            DrawText(myText, 0, 0, MyBase.Bounds.Width, MyBase.Bounds.Height, myColor)

        End Sub

        Private myDuration As TimeSpan = TimeSpan.Zero

        Public Overrides Sub Tick(ByVal delta As TimeSpan)
            myDuration += delta

            If myDuration >= myTimeSpan Then
                EndOfAction()
            End If

        End Sub
    End Class

    Private myHugeTexts As New List(Of ActionHugeText)()

    Private Sub RemoveHugeText(ByVal obj As Action)
        myHugeTexts.Remove(DirectCast(obj, ActionHugeText))
    End Sub

    Private Class QElement
        Public Texture As TextureInfo
        Public Type As Integer
        Public Src As Rectangle
        Public Bounds As New List(Of Rectangle)(1)
    End Class

    Private Class ObjectInfo
        Public E As QElement
        Public X As Integer
        Public Y As Integer
        Public FX(4) As Integer
    End Class

    Private geometry()() As Byte = Nothing

    Private viewCenter As New Point(0, 300)

    Public Sub New(ByVal unreg As UnregisterActionCallback, ByVal parent As ScriptEngine, ByVal script As ModuleInfo, ByVal args() As ScriptValue)
        MyBase.New(unreg, parent, script, args)

        Dim lines As New List(Of Byte())
        Using reader As New IO.StreamReader(IO.File.OpenRead("SuperQwerty\geo1.txt"))
            Dim line As String = reader.ReadLine()
            Do While Not line Is Nothing
                With System.Text.RegularExpressions.Regex.Match(line, "(\w\w)(\w\w)(\w\w)(\w\w)(\w\w)(\w\w)(\w\w)(\w\w)(\w\w)(\w\w)(\w\w)(\w\w)(\w\w)(\w\w)(\w\w)")
                    If .Success Then
                        Dim bt(14) As Byte
                        For i As Integer = 0 To bt.Length - 1
                            bt(i) = [Byte].Parse(.Groups(i + 1).Value)
                        Next i
                        lines.Add(bt)
                    End If
                End With
                line = reader.ReadLine()
            Loop
        End Using
        lines.Reverse()

        geometry = lines.ToArray()
        Call MakeGeoMap()

    End Sub

    Private myElementDic As New Dictionary(Of String, QElement)(16)
    Private myObjectDic As New Dictionary(Of Integer, ObjectInfo)(256)

    Private Function cmdAddElement(ByVal params() As ScriptValue) As ScriptValue
        Dim info As New QElement()
        myElementDic.Add(params(0).AsString(), info)
        info.Texture = TextureArray.Global(params(1).AsString())
        info.Type = params(2).AsInt32()
        info.Src = New Rectangle(Point.Empty, info.Texture.Size)
        info.Bounds.Add(info.Src)
        Return 0
    End Function

    Private Function cmdAddElementEx(ByVal params() As ScriptValue) As ScriptValue
        Dim info As New QElement()
        myElementDic.Add(params(0).AsString(), info)
        info.Texture = TextureArray.Global(params(1).AsString())
        info.Type = params(2).AsInt32()
        info.Src = New Rectangle(params(3).AsInt32(), params(4).AsInt32(), params(5).AsInt32(), params(6).AsInt32())
        For i As Integer = 7 To params.Length - 1 Step 4
            info.Bounds.Add(New Rectangle(params(i).AsInt32(), params(i + 1).AsInt32(), params(i + 2).AsInt32(), params(i + 3).AsInt32()))
        Next i
        Return 0
    End Function

    Private Function cmdGetObjectWidth(ByVal params() As ScriptValue) As ScriptValue
        Return myObjectDic(params(0).AsInt32()).E.Src.Width
    End Function

    Private Function cmdGetObjectHeight(ByVal params() As ScriptValue) As ScriptValue
        Return myObjectDic(params(0).AsInt32()).E.Src.Height
    End Function

    Private Function cmdAddObject(ByVal params() As ScriptValue) As ScriptValue
        Dim info As New ObjectInfo()
        Dim id As Integer = info.GetHashCode()
        myObjectDic.Add(id, info)
        info.E = myElementDic(params(0).AsString())
        info.X = params(1).AsInt32() + 400
        info.Y = params(2).AsInt32()
        Return id
    End Function

    Private Function cmdRemoveObject(ByVal params() As ScriptValue) As ScriptValue
        myObjectDic.Remove(params(0).AsInt32())
        Return 0
    End Function

    Private Function cmdGetObjectType(ByVal params() As ScriptValue) As ScriptValue
        Return myObjectDic(params(0).AsInt32()).E.Type
    End Function

    Private Function cmdGetCollidedObjects(ByVal params() As ScriptValue) As ScriptValue
        Dim targets(params.Length - 2) As Integer
        For i As Integer = 0 To targets.Length - 1
            targets(i) = params(i + 1).AsInt32()
        Next i
        Dim theList As New List(Of ScriptValue)(myObjectDic.Count)
        Dim thisObject As ObjectInfo = myObjectDic(params(0).AsInt32())
        For Each tb As Rectangle In thisObject.E.Bounds
            Dim thisBounds As New Rectangle(New Point(thisObject.X + tb.X, thisObject.Y + tb.Y), tb.Size)
            For Each kv As KeyValuePair(Of Integer, ObjectInfo) In myObjectDic
                Dim obj As ObjectInfo = kv.Value
                For i As Integer = 0 To targets.Length - 1
                    If kv.Value.E.Type = targets(i) Then
                        For Each b As Rectangle In obj.E.Bounds
                            If thisBounds.IntersectsWith(New Rectangle(New Point(obj.X + b.X, obj.Y + b.Y), b.Size)) Then
                                theList.Add(kv.Key)
                                Exit For
                            End If
                        Next b
                    End If
                Next i
            Next kv
        Next tb
        Return New ScriptValue(theList)
    End Function

    Private Function cmdSetPos(ByVal params() As ScriptValue) As ScriptValue
        With myObjectDic(params(0).AsInt32())
            .X = params(1).AsInt32() + 400
            .Y = params(2).AsInt32()
        End With
        Return 0
    End Function

    Private Sub MakeGeoMap()

        bgpMap.Clear()
        Dim viewBounds As New Rectangle(viewCenter.X - 240, viewCenter.Y - 300, 480, 600)
        Dim shift As New Point(400 - viewCenter.X, 300 - viewCenter.Y)
        For i As Integer = 0 To geometry.Length - 1
            Dim theY As Integer = i * 32
            For j As Integer = 0 To geometry(i).Length - 1
                Dim theBounds As New Rectangle(j * 32 - 240, theY, 32, 32)
                If viewBounds.IntersectsWith(theBounds) Then
                    Dim geo As Byte = geometry(i)(j)
                    Dim theList As List(Of Point) = Nothing
                    If Not bgpMap.TryGetValue(geo, theList) Then
                        theList = New List(Of Point)(128)
                        bgpMap.Add(geo, theList)
                    End If
                    theList.Add(New Point(theBounds.X + shift.X, 600 - (theBounds.Y + shift.Y) - 32))
                End If
            Next j
        Next i

    End Sub

    Private Function cmdScrollBG(ByVal params() As ScriptValue) As ScriptValue
        Dim x As Integer = params(0).AsInt32()
        Dim y As Integer = params(1).AsInt32()
        Dim oc As Point = viewCenter
        viewCenter.Offset(x, y)
        If x <> 0 OrElse y <> 0 Then
            Call MakeGeoMap()
        End If
        Return 0
    End Function

    Private myTimer As New Stopwatch()

    Private Function cmdStartTimer(ByVal params() As ScriptValue) As ScriptValue
        myTimer.Start()
        Return 0
    End Function

    Private Function cmdStopTimer(ByVal params() As ScriptValue) As ScriptValue
        myTimer.Stop()
        Return 0
    End Function

    Private Function cmdResetTimer(ByVal params() As ScriptValue) As ScriptValue
        myTimer.Reset()
        Return 0
    End Function

    Private Function cmdGetTime(ByVal params() As ScriptValue) As ScriptValue
        Return CInt(myTimer.ElapsedMilliseconds)
    End Function

    Private Function cmdSetFx(ByVal params() As ScriptValue) As ScriptValue
        With myObjectDic(params(0).AsInt32())
            For i As Integer = 1 To params.Length - 1
                .FX(i - 1) = params(i).AsInt32()
            Next i
        End With
        Return 0
    End Function

    Private Function cmdSetGageR(ByVal params() As ScriptValue) As ScriptValue
        myGagePowerR = params(0).AsInt32()
        Return 0
    End Function

    Private Function cmdAddText(ByVal params() As ScriptValue) As ScriptValue
        myHugeTexts.Add(New ActionHugeText(AddressOf Me.RemoveHugeText, New Rectangle(params(4).AsInt32(), params(5).AsInt32(), params(6).AsInt32(), params(7).AsInt32()), _
            New Font("MS GOTHIC", params(1).AsInt32(), FontStyle.Bold), params(0).AsString(), Color.FromArgb(params(2).AsInt32()), TimeSpan.FromMilliseconds(params(3).AsInt32())))
        Return 0
    End Function

    Private Function cmdSin(ByVal params() As ScriptValue) As ScriptValue
        Return Math.Sin(params(0).AsDouble())
    End Function

    Private Function cmdGoToLevel(ByVal params() As ScriptValue) As ScriptValue
        myTimer.Stop()
        myTimer.Reset()
        myElementDic.Clear()
        myObjectDic.Clear()
        For Each aht As ActionHugeText In myHugeTexts.ToArray()
            aht.EndOfAction()
        Next aht
        viewCenter = New Point(0, 300)
        Return 0
    End Function

    Protected Overrides Function GetCommandMap() As IDictionary(Of String, CommandImpl)

        Dim dic As New Dictionary(Of String, CommandImpl)(MyBase.GetCommandMap())

        dic.Add("AddElement", AddressOf cmdAddElement)
        dic.Add("AddElementEx", AddressOf cmdAddElementEx)
        dic.Add("AddObject", AddressOf cmdAddObject)
        dic.Add("RemoveObject", AddressOf cmdRemoveObject)
        dic.Add("GetCollidedObjects", AddressOf cmdGetCollidedObjects)
        dic.Add("SetPos", AddressOf cmdSetPos)
        dic.Add("StartTimer", AddressOf cmdStartTimer)
        dic.Add("StopTimer", AddressOf cmdStopTimer)
        dic.Add("ResetTimer", AddressOf cmdResetTimer)
        dic.Add("GetTime", AddressOf cmdGetTime)
        dic.Add("ScrollBG", AddressOf cmdScrollBG)
        dic.Add("GetObjectType", AddressOf cmdGetObjectType)
        dic.Add("SetFx", AddressOf cmdSetFx)
        dic.Add("SetGageR", AddressOf cmdSetGageR)
        dic.Add("AddText", AddressOf cmdAddText)
        dic.Add("Sin", AddressOf cmdSin)
        dic.Add("GetObjectWidth", AddressOf cmdGetObjectWidth)
        dic.Add("GetObjectHeight", AddressOf cmdGetObjectHeight)
        dic.Add("GoToLevel", AddressOf cmdGoToLevel)

        Return dic

    End Function

    Dim bgpMap As New Dictionary(Of Byte, List(Of Point))

    Private myGagePowerR As Integer = 0

    Public Overrides Sub Render(ByVal r As IGraphicsDevice)
        MyBase.Render(r)

        With CType(r, Renderer)
            .Fill(Color.Blue)

            .Paint(New Rectangle(660, 320, 50, 260), Color.Black)
            Dim rh As Integer = myGagePowerR * 260 \ 100
            .Paint(New Rectangle(660, 320 + (260 - rh), 50, rh), Color.FromArgb(myGagePowerR * 255 \ 100, 0, 0))

            .SetTexture(TextureArray.Global("QGeo"))
            Try
                Dim xlst As New List(Of Single)(256)
                Dim ylst As New List(Of Single)(256)
                For Each kv As KeyValuePair(Of Byte, List(Of Point)) In bgpMap
                    For Each p As Point In kv.Value
                        xlst.Add(p.X)
                        ylst.Add(p.Y)
                    Next p
                    .RenderBatch(xlst.ToArray(), ylst.ToArray(), 32, 32, (kv.Key Mod 8) * 32, (kv.Key \ 8) * 32)
                    xlst.Clear()
                    ylst.Clear()
                Next kv
            Finally
                .ResetTexture()
            End Try

            For Each aht As ActionHugeText In myHugeTexts
                aht.Render(r)
            Next aht

            For Each kv As KeyValuePair(Of Integer, ObjectInfo) In myObjectDic
                Dim obj As ObjectInfo = kv.Value
                .SetTexture(kv.Value.E.Texture)
                Try
                    Select Case kv.Value.FX(0)
                        Case 0
                            .Render(obj.X, obj.Y, obj.E.Src.Width, obj.E.Src.Height, obj.E.Src.X, obj.E.Src.Y)
                        Case 1
                            .Render(obj.X, obj.Y, obj.E.Src.Width, obj.E.Src.Height, obj.E.Src.X, obj.E.Src.Y, kv.Value.FX(1))
                        Case 2
                            .Render(obj.X + obj.FX(1), kv.Value.Y + kv.Value.FX(2), kv.Value.FX(3), kv.Value.FX(4), kv.Value.FX(1), kv.Value.FX(2))
                    End Select
                Finally
                    .ResetTexture()
                End Try
            Next kv

        End With

        MyBase.RenderAllProcess(r)


    End Sub

    Public Overrides Sub Tick(ByVal delta As TimeSpan)

        For Each aht As ActionHugeText In myHugeTexts.ToArray()
            aht.Tick(delta)
        Next aht

        TickAllProcess(delta)
        MyBase.Tick(delta)

    End Sub
End Class

Public Class QwertyStarter
    Implements IHasCommand

    Private myTarget As ScriptEngine = Nothing

    Private Function cmdEnterTheQwerty(ByVal params() As ScriptValue) As ScriptValue
        myTarget.RegisterThreadAction(New SuperQwerty(AddressOf myTarget.UnregisterThreadAction, myTarget, _
            BWV1048.Johan_Sebastian_Bach.ParseModule("superqwerty\level1"), New ScriptValue() {}))
        Return 0
    End Function

    Public Function GetCommandMap() As IDictionary(Of String, CommandImpl) Implements IHasCommand.GetCommandMap

        Dim dic As New Dictionary(Of String, CommandImpl)

        dic.Add("EnterTheQwerty", AddressOf cmdEnterTheQwerty)

        Return dic

    End Function

    Public Sub SetTarget(ByVal obj As ScriptEngine) Implements IHasCommand.SetTarget

        myTarget = obj

    End Sub
End Class
